# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import pytest

from radish.stepmodel import Step
from radish.exceptions import RadishError


def test_creating_simple_step():
    """
    Test creating a simple Step
    """
    # given & when
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )

    # then
    assert step.id == 1
    assert step.sentence == "I am a Step"
    assert step.path == "foo.feature"
    assert step.line == 1
    assert step.parent is None
    assert step.runable is True
    assert step.context_class is None
    assert step.embeddings == []


def test_getting_step_context_object(mocker):
    """
    Test getting a Step's context object
    """
    # given
    scenario = mocker.MagicMock(context=42)
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=scenario,
        runable=True,
        context_class=None,
    )

    # when
    context = step.context

    # then
    assert context == 42
    assert context == scenario.context


def test_getting_expanded_sentence(mocker):
    """
    Test getting the expanded Step sentence
    """
    # given
    scenario = mocker.MagicMock(constants=[("foo", "42"), ("bar", "21"), ("bla", "33")])
    step = Step(
        1,
        "I am ${foo} and ${bar} bla",
        "foo.feature",
        1,
        parent=scenario,
        runable=True,
        context_class=None,
    )

    # when
    sentence = step.expanded_sentence

    # then
    assert sentence == "I am 42 and 21 bla"
    assert step.sentence == "I am ${foo} and ${bar} bla"


def test_getting_context_sensitive_sentence(mocker):
    """
    Test getting the context sensitive Step sentence
    """
    # given
    scenario = mocker.MagicMock(constants=[("foo", "42"), ("bar", "21"), ("bla", "33")])
    step = Step(
        1,
        "And I am ${foo} and ${bar} bla",
        "foo.feature",
        1,
        parent=scenario,
        runable=True,
        context_class="Given",
    )
    step_no_context = Step(
        1,
        "And I am ${foo} and ${bar} bla",
        "foo.feature",
        1,
        parent=scenario,
        runable=True,
        context_class=None,
    )

    # when
    sentence = step.context_sensitive_sentence
    sentence_no_context = step_no_context.context_sensitive_sentence

    # then
    assert sentence == "Given I am 42 and 21 bla"
    assert step.sentence == "And I am ${foo} and ${bar} bla"

    assert sentence_no_context == "And I am 42 and 21 bla"
    assert step_no_context.sentence == "And I am ${foo} and ${bar} bla"


@pytest.mark.parametrize(
    "given_raw_text, expected_text",
    [
        ([], ""),
        (["Hello", "World"], "Hello\nWorld"),
        (["Hello", "Awesome", "World"], "Hello\nAwesome\nWorld"),
    ],
)
def test_getting_raw_text_from_step(given_raw_text, expected_text):
    """
    Test getting a Step's raw text data
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )
    step.raw_text = given_raw_text

    # when
    text = step.text

    # then
    assert text == expected_text


@pytest.mark.parametrize("debug_or_run", [("debug"), ("run")])
def test_run_not_runable_step(debug_or_run, mock_utils_debugger):
    """
    Test running a non-runable Step
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=False,
        context_class=None,
    )

    # when
    method = getattr(step, debug_or_run)
    state = method()

    # then
    assert state == Step.State.UNTESTED


def test_run_step_with_invalid_defintion_func():
    """
    Test running a Step with an invalid definition function
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )

    # when the Step doesn't have a definition function
    with pytest.raises(RadishError) as exc:
        step.run()
    # then the Step fails with an Exception
    assert str(exc.value) == "The step 'I am a Step' does not have a step definition"

    # when the Step has a non-callable definition function
    step.definition_func = "not-callable"
    with pytest.raises(RadishError) as exc:
        step.run()
    # then the step fails with an Exception
    assert str(exc.value) == "The step 'I am a Step' does not have a step definition"


class StepHelper(object):
    """Helper class only used for mocking"""

    @staticmethod
    def step_func(step, foo=None, bar=None):
        """
        Helper Step Definition Function.
        """
        pass

    @staticmethod
    def step_pending_func(step):
        """
        Helper Step Definition Function which
        sets the Step to PENDING
        """
        step.pending()

    @staticmethod
    def step_fail_func(step):
        """
        Helper Step Definition Function which fails the Step
        """
        raise AssertionError("failing step")

    @staticmethod
    def step_skip_func(step):
        """
        Helper Step Definition Function which skips the step
        """
        step.skip()

@pytest.mark.parametrize("debug_or_run", [("run"), ("debug")])
def test_run_debug_step_function_with_kwargs(debug_or_run, mocker, mock_utils_debugger):
    """
    Test running/debugging a Step with a function and keyword arguments
    """
    # mock step function which is to use
    mocker.spy(StepHelper, "step_func")

    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )
    step.definition_func = StepHelper.step_func
    step.argument_match = mocker.MagicMock()
    step.argument_match.evaluate.return_value = (tuple(), {"foo": "1", "bar": "2"})

    # when
    method = getattr(step, debug_or_run)
    state = method()

    # then
    assert state == Step.State.PASSED
    StepHelper.step_func.assert_called_once_with(step, foo="1", bar="2")


@pytest.mark.parametrize("debug_or_run", [("run"), ("debug")])
def test_run_debug_step_function_with_posargs(
    debug_or_run, mocker, mock_utils_debugger
):
    """
    Test running/debugging a Step with a function and positional arguments
    """
    # mock step function which is to use
    mocker.spy(StepHelper, "step_func")

    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )
    step.definition_func = StepHelper.step_func
    step.argument_match = mocker.MagicMock()
    step.argument_match.evaluate.return_value = ((1, 2), {})

    # when
    method = getattr(step, debug_or_run)
    state = method()

    # then
    assert state == Step.State.PASSED
    StepHelper.step_func.assert_called_once_with(step, 1, 2)


@pytest.mark.parametrize("debug_or_run", [("run"), ("debug")])
def test_run_debug_step_function_mark_pending(
    debug_or_run, mocker, mock_utils_debugger
):
    """
    Test running/debugging a Step which marks itself as pending
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )
    step.definition_func = StepHelper.step_pending_func
    step.argument_match = mocker.MagicMock()
    step.argument_match.evaluate.return_value = (tuple(), {})

    # when
    method = getattr(step, debug_or_run)
    state = method()

    # then
    assert state == Step.State.PENDING == step.state


@pytest.mark.parametrize("debug_or_run", [("run"), ("debug")])
def test_run_debug_step_function_mark_skipped(
    debug_or_run, mocker, mock_utils_debugger
):
    """
    Test running/debugging a Step which marks itself as skipped
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )
    step.definition_func = StepHelper.step_skip_func
    step.argument_match = mocker.MagicMock()
    step.argument_match.evaluate.return_value = (tuple(), {})

    # when
    method = getattr(step, debug_or_run)
    state = method()

    # then
    assert state == Step.State.SKIPPED == step.state


@pytest.mark.parametrize("debug_or_run", [("run"), ("debug")])
def test_run_debug_step_function_with_exception(
    debug_or_run, mocker, mock_utils_debugger
):
    """
    Test running/debugging a Step which raises an Exception
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )
    step.definition_func = StepHelper.step_fail_func
    step.argument_match = mocker.MagicMock()
    step.argument_match.evaluate.return_value = (tuple(), {})

    # when
    method = getattr(step, debug_or_run)
    state = method()

    # then
    assert state == step.state == Step.State.FAILED
    assert step.failure is not None
    assert step.failure.reason == "failing step"
    assert step.failure.name == "AssertionError"


def test_skip_a_step():
    """
    Test skipping a Step
    """
    # given
    step = Step(
        1,
        "I am a Step",
        "foo.feature",
        1,
        parent=None,
        runable=True,
        context_class=None,
    )

    # when
    step.skip()

    # then
    assert step.state == Step.State.SKIPPED



# -*- coding: utf-8 -*-

from radish import custom_type


@custom_type("is_plugged_in", r"(plugged in|not plugged in)")
def is_plugged_in(text):
    return text == "plugged in"


@custom_type("is_registered", r"(not )?to be registered")
def is_registered(text):
    return text == "to be registered"



# -*- coding: utf-8 -*-

"""
    This module provides a Scenario type which represents a Scenario loop
"""

from __future__ import unicode_literals

import copy

from .scenario import Scenario
from .iterationscenario import IterationScenario
from .stepmodel import Step


class ScenarioLoop(Scenario):
    """
        Represents a scenario loop
    """

    def __init__(
        self,
        id,
        keyword,
        iterations_keyword,
        sentence,
        path,
        line,
        parent,
        tags=None,
        preconditions=None,
        background=None,
    ):
        super(ScenarioLoop, self).__init__(
            id,
            keyword,
            sentence,
            path,
            line,
            parent,
            tags,
            preconditions,
            background=background,
        )
        self.iterations_keyword = iterations_keyword
        self.iterations = 0
        self.scenarios = []

    def build_scenarios(self):
        """
            Builds the scenarios for every iteration

            Note: This must be done before merging the steps
                  from the feature file with the step definitions
        """
        for i in range(self.iterations):
            scenario_id = self.id + i + 1
            background = None
            scenario = IterationScenario(
                scenario_id,
                self.keyword,
                "{0} - iteration {1}".format(self.sentence, i),
                self.path,
                self.line,
                self,
                i,
            )
            if self.background:
                background = self.background.create_instance(
                    parent=scenario, steps_runable=True
                )
                scenario.background = background

            for step_id, iteration_step in enumerate(self.steps):
                step = Step(
                    step_id + 1,
                    iteration_step.sentence,
                    iteration_step.path,
                    iteration_step.line,
                    scenario,
                    True,
                    context_class=iteration_step.context_class,
                )
                step.table_header = copy.copy(iteration_step.table_header)
                step.table_data = copy.copy(iteration_step.table_data)
                step.table = copy.copy(iteration_step.table)
                step.raw_text = copy.copy(iteration_step.raw_text)
                scenario.steps.append(step)
            self.scenarios.append(scenario)

    def after_parse(self):
        """
            Build looped scenarios
        """
        Scenario.after_parse(self)
        self.build_scenarios()
        self.complete = True



# -*- coding: utf-8 -*-

"""
    This module provides several utility functions
"""

from __future__ import unicode_literals
from __future__ import print_function

import os
import re
import sys
import fnmatch
import traceback
import warnings
import pydoc
import itertools
import calendar
from datetime import datetime, timedelta

from .compat import u


class Failure(object):  # pylint: disable=too-few-public-methods
    """
        Represents the fail reason for a step
    """

    def __init__(self, exception):
        """
            Initalizes the Step failure with a given Exception

            :param Exception exception: the exception shrown in the step
        """
        self.exception = exception
        self.reason = u(str(exception))
        self.traceback = u(traceback.format_exc())
        self.name = exception.__class__.__name__
        traceback_info = traceback.extract_tb(sys.exc_info()[2])[-1]
        self.filename = traceback_info[0]
        self.line = int(traceback_info[1])


def console_write(text):
    """
        Writes the given text to the console

        :param str text: the text which is printed to the console
    """
    print(u(text))


def expandpath(path):
    """
        Expands a path

        :param string path: the path to expand
    """
    return os.path.expanduser(os.path.expandvars(path))


def recursive_glob(root, pattern):
    """
        Recursively search for files with given pattern inside a path

        :param str root: the root location to start search
        :param str pattern: to pattern to look for. It's matched against the filenames under `root`.

        :rtype: list
        :returns: A list of matching files
    """
    matches = []
    for root, dirnames, filenames in os.walk(root):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches


def get_debugger():
    """
        Returns a debugger instance
    """
    try:
        from IPython.core.debugger import Pdb

        pdb = Pdb()
    except ImportError:
        try:
            from IPython.Debugger import Pdb
            from IPython.Shell import IPShell

            IPShell(argv=[""])
            pdb = Pdb()
        except ImportError:
            warnings.warn(
                'pdb was selected as a debugger. If you want to use ipython as a debugger you have to "pip install radish-bdd[ipython-debugger]"'
            )
            import pdb

    return pdb


def format_utc_to_local_tz(utc_dt, fmt="%Y-%m-%dT%H:%M:%S"):
    """
    Formats the given UTC datetime as a string converted to the local timezone.

    If the datetime is ``None``, it will return an empty string.

    Args:
        datetime (Datetime): the UTC based datetime object
        fmt (str): optional format for the string
    """
    if not utc_dt:
        return ""

    def utc_to_local(utc_dt):
        timestamp = calendar.timegm(utc_dt.timetuple())
        local_dt = datetime.fromtimestamp(timestamp)
        assert utc_dt.resolution >= timedelta(microseconds=1)
        return local_dt.replace(microsecond=utc_dt.microsecond)

    return utc_to_local(utc_dt).strftime(fmt)


def make_unique_obj_list(somelist, attr):
    """
        Make list with objects unique
        according to an objects attribute.
    """
    tmp = {}
    for item in somelist:
        tmp[attr(item)] = item
    return tmp.values()


def get_func_code(func):
    """
        Get the code object for the given function.
    """
    if sys.version_info[0] == 3:
        return func.__code__
    else:
        return func.func_code


def get_func_arg_names(func):
    """
        Get the argument names of the given function.
    """
    return get_func_code(func).co_varnames


def get_func_location(func):
    """
        Get the location where the given function is implemented.
    """
    func_code = get_func_code(func)
    return "{0}:{1}".format(func_code.co_filename, func_code.co_firstlineno)


def str_lreplace(pattern, replacement, string, escape_pattern=False, flags=0):
    """
    Only replace the pattern with replacement
    if the string starts with the pattern.
    """
    if escape_pattern:
        pattern = re.escape(pattern)

    return re.sub(r"^{0}".format(pattern), replacement, string, flags=flags)


def locate(name):
    """
    Locate the object for the given name
    """
    obj = pydoc.locate(name)
    if not obj:
        obj = globals().get(name, None)

    return obj


def flattened_basedirs(basedirs):
    """
    Flatten a list of basedirs.

    Multiple basedirs can be specified within a
    single element split by a colon.
    """
    separator = ";" if os.name == "nt" else ":"
    return list(x for x in itertools.chain(*(x.split(separator) for x in basedirs)) if x)


def split_unescape(s, delim, escape='\\', unescape=True):
    """
    >>> split_unescape('foo|bar', '|')
    ['foo', 'bar']
    >>> split_unescape(r'foo\|bar', '|')
    ['foo|bar']
    >>> split_unescape(r'foo\\|bar', '|', unescape=True)
    [r'foo|', 'bar']
    >>> split_unescape(r'foo\\|bar', '|', unescape=False)
    [r'foo\\', 'bar']
    >>> split_unescape(r'foo\', '|', unescape=True)
    [r'foo\']
    """
    ret = []
    current = []
    itr = iter(s)
    for ch in itr:
        if ch == escape:
            try:
                # skip the next character; it has been escaped!
                if not unescape:
                    current.append(escape)
                current.append(next(itr))
            except StopIteration:
                if unescape:
                    current.append(escape)
        elif ch == delim:
            # split! (add current to the list and reset it)
            ret.append(''.join(current))
            current = []
        else:
            current.append(ch)
    ret.append(''.join(current))
    return ret



# -*- coding: utf-8 -*-

"""
    This module provides some functionality to diagnose thrown exceptions
"""

from __future__ import unicode_literals


import sys
from functools import wraps

import colorful

from .terrain import world
from .exceptions import (
    RadishError,
    FeatureFileSyntaxError,
    StepDefinitionNotFoundError,
    HookError,
    SameStepError,
)
from .utils import Failure, console_write


__RADISH_DOC__ = "https://github.com/radish-bdd/radish"


def write_error(text):
    """
        Writes the given text to the console
    """
    console_write("{0}: {1}".format(colorful.bold_red("Error"), colorful.red(text)))


def write_failure(failure):
    """
        Writes the failure to the console
    """
    console_write("\n{0}".format(colorful.red(failure.traceback)))


def abort(return_code):
    """
        Aborts the program with the given return_code
    """
    sys.exit(return_code)


def error_oracle(func):
    """
        Decorator to diagnose thrown exceptions
    """

    @wraps(func)
    def _decorator(*args, **kwargs):
        """
            The actual decorator
        """
        try:
            return func(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            handle_exception(e)

    return _decorator


def catch_unhandled_exception(exc_type, exc_value, traceback):
    """
        Catch all unhandled exceptions
    """
    handle_exception(exc_value)


def handle_exception(exception):
    """
        Handle the given exception

        This will print more information about the given exception

        :param Exception exception: the exception to handle
    """
    if isinstance(exception, HookError):
        write_error(exception)
        write_failure(exception.failure)
        abort(1)
    elif isinstance(exception, RadishError):
        write_error(exception)
        abort(1)
    elif isinstance(exception, KeyboardInterrupt):
        console_write("Aborted by the user...")
        abort(1)
    else:
        write_error(exception)
        write_failure(Failure(exception))
        abort(2)



# -*- coding: utf-8 -*-

from radish import step, given, when, then


@step("a user named {:w}")
def have_user(step, username):
    # print(step.context)
    step.context.users[username] = {"site": None}


@step("a personal site owned by {username:w}")
def have_personal_site(step, username):
    # print(step.context)
    step.context.users[username]["site"] = {"access": []}


@step("{username:w} grants access to {consumer:w}")
def grant_personal_site_access(step, username, consumer):
    # print(step.context)
    step.context.users[username]["site"]["access"].append(consumer)


@when("I'm logged in as {username:w}")
def logged_in(step, username):
    # print(step.context)
    step.context.current_user = username


@then("I can access {username:w} personal site")
def can_access_personal_site(step, username):
    # print(step.context)
    assert step.context.current_user in step.context.users[username]["site"]["access"]


@then("I cannot access {username:w} personal site")
def cannot_access_personal_site(step, username):
    # print(step.context)
    assert (
        step.context.current_user not in step.context.users[username]["site"]["access"]
    )



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""


from radish.background import Background
from radish.stepmodel import Step


def test_creating_simple_background():
    """
    Test creating a simple Background
    """
    # given & when
    background = Background(
        "Background", "I am a Background", "foo.feature", 1, parent=None
    )

    # then
    assert background.id is None
    assert background.keyword == "Background"
    assert background.sentence == "I am a Background"
    assert background.path == "foo.feature"
    assert background.line == 1
    assert background.parent is None


def test_creating_a_concrete_background_instance():
    """
    Test creating a concrete Background instance
    """
    # given & when
    background = Background(
        "Background", "I am a Background", "foo.feature", 1, parent=None
    )
    # add some Steps
    background.steps.append(Step(1, "Foo", "foo.feature", 2, background, False))
    background.steps.append(Step(2, "Foo", "foo.feature", 3, background, False))

    # when
    instance = background.create_instance()

    # then
    assert len(instance.steps) == 2
    assert background.steps[0] != instance.steps[0]
    assert background.steps[1] != instance.steps[1]



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

from __future__ import unicode_literals

import re

import pytest

from radish.stepregistry import step, steps
from radish.stepregistry import given, when, then
from radish.compat import PY2, re_pattern
import radish.exceptions as errors


def test_registering_simple_steps(stepregistry):
    """
    Test registering simple Step functions
    """
    # given
    def step_a():
        pass

    def step_b():
        pass

    # when
    stepregistry.register("step_pattern_a", step_a)
    stepregistry.register("step_pattern_b", step_b)

    # then
    assert len(stepregistry.steps) == 2
    assert stepregistry.steps["step_pattern_a"] == step_a
    assert stepregistry.steps["step_pattern_b"] == step_b


def test_registering_step_pattern_twice(stepregistry):
    """
    Test registering the same Step pattern twice
    """
    # given
    def step_a():
        pass

    def step_b():
        pass

    stepregistry.register("step_pattern_a", step_a)

    # when
    with pytest.raises(errors.SameStepError) as exc:
        stepregistry.register("step_pattern_a", step_b)

    # then
    assert str(exc.value).startswith(
        "Cannot register step step_b with regex 'step_pattern_a' because it is already used by step step_a"
    )


def test_registering_steps_via_object(stepregistry):
    """
    Test registering Steps via object
    """
    # given
    class MySteps(object):
        def some_step(self):
            """When I call some step"""

        def some_other_step(self):
            """
            I do some stuff

            This is not part of the step pattern
            """

    # when
    steps_object = MySteps()
    stepregistry.register_object(steps_object)

    # then
    assert len(stepregistry.steps) == 2
    assert stepregistry.steps["When I call some step"] == steps_object.some_step
    assert stepregistry.steps["I do some stuff"] == steps_object.some_other_step


def test_ignore_methods_registering_object(stepregistry):
    """
    Test ignoring methods when registering an object
    """
    # given
    class MySteps(object):
        ignore = ["some_method"]

        def some_step(self):
            """When I call some step"""

        def some_method(self):
            pass

    # when
    steps_object = MySteps()
    stepregistry.register_object(steps_object)

    # then
    assert len(stepregistry.steps) == 1
    assert stepregistry.steps["When I call some step"] == steps_object.some_step


def test_error_if_no_step_regex_given_for_object(stepregistry):
    """
    Test error if a step method in object has no regex
    """
    # given
    class MySteps(object):
        def some_step(self):
            pass

    # when
    steps_object = MySteps()

    with pytest.raises(errors.RadishError) as exc:
        stepregistry.register_object(steps_object)

    # then
    assert (
        str(exc.value)
        == "Step definition 'some_step' from class must have step regex in docstring"
    )


def test_invalid_regex_step_pattern_in_method_docstring(stepregistry):
    """
    Test invalid regex Step pattern in method docstring
    """
    # given
    class MySteps(object):
        def some_step(self):
            """
            So (( invalid )(
            """

    # when
    steps_object = MySteps()

    with pytest.raises(errors.StepRegexError) as exc:
        stepregistry.register_object(steps_object)

    # then
    assert str(exc.value).startswith(
        "Cannot compile regex 'So (( invalid )(' from step"
    )


@pytest.mark.parametrize(
    "pattern",
    ["I do some stuff", re.compile("I do some stuff")],
    ids=["Step with Step Pattern", "Step with Regex"],
)
def test_registering_step_function_via_step_decorator(pattern, stepregistry):
    """
    Test registering Step function via step decorator
    """
    # given & when
    @step(pattern)
    def step_a(step):
        pass

    # then
    assert len(stepregistry.steps) == 1
    if not PY2 or not isinstance(pattern, re_pattern):  # doesn't work for Python 2 and re_pattern.
        assert stepregistry.steps[pattern] == step_a


@pytest.mark.parametrize(
    "pattern, expected_pattern",
    [
        ("I do some stuff", "Given I do some stuff"),
        (re.compile("I do some stuff"), re.compile("Given I do some stuff")),
    ],
    ids=["Step with Step Pattern", "Step with Regex"],
)
def test_registering_step_function_via_given_decorator(
    pattern, expected_pattern, stepregistry
):
    """
    Test registering Step function via given decorator
    """
    # given & when
    @given(pattern)
    def step_a(step):
        pass

    # then
    assert len(stepregistry.steps) == 1
    if not PY2 or not isinstance(pattern, re_pattern):  # doesn't work for Python 2 and re_pattern.
        assert stepregistry.steps[expected_pattern] == step_a


@pytest.mark.parametrize(
    "pattern, expected_pattern",
    [
        ("I do some stuff", "When I do some stuff"),
        (re.compile("I do some stuff"), re.compile("When I do some stuff")),
    ],
    ids=["Step with Step Pattern", "Step with Regex"],
)
def test_registering_step_function_via_when_decorator(
    pattern, expected_pattern, stepregistry
):
    """
    Test registering Step function via when decorator
    """
    # given & when
    @when(pattern)
    def step_a(step):
        pass

    # then
    assert len(stepregistry.steps) == 1
    if not PY2 or not isinstance(pattern, re_pattern):  # doesn't work for Python 2 and re_pattern.
        assert stepregistry.steps[expected_pattern] == step_a


@pytest.mark.parametrize(
    "pattern, expected_pattern",
    [
        ("I do some stuff", "Then I do some stuff"),
        (re.compile("I do some stuff"), re.compile("Then I do some stuff")),
    ],
    ids=["Step with Step Pattern", "Step with Regex"],
)
def test_registering_step_function_via_then_decorator(
    pattern, expected_pattern, stepregistry
):
    """
    Test registering Step function via then decorator
    """
    # given & when
    @then(pattern)
    def step_a(step):
        pass

    # then
    assert len(stepregistry.steps) == 1
    if not PY2 or not isinstance(pattern, re_pattern):  # doesn't work for Python 2 and re_pattern.
        assert stepregistry.steps[expected_pattern] == step_a


def test_registering_steps_from_object_via_steps_decorator(stepregistry):
    """
    Test registering Steps from object via steps decorator
    """
    # given & when
    @steps
    class MySteps(object):
        def some_step(self):
            """When I call some step"""

        def some_other_step(self):
            """
            I do some stuff

            This is not part of the step pattern
            """

    # then
    assert len(stepregistry.steps) == 2
    assert "When I call some step" in stepregistry.steps
    assert "I do some stuff" in stepregistry.steps


def test_getting_pattern_of_specific_func(stepregistry):
    """
    Test getting the pattern of a specific func
    """
    # given
    def step_a():
        pass

    def step_b():
        pass

    def step_c():
        pass

    stepregistry.register("step_pattern_a", step_a)
    stepregistry.register("step_pattern_b", step_b)

    # when
    pattern_a = stepregistry.get_pattern(step_a)
    pattern_b = stepregistry.get_pattern(step_b)
    pattern_c = stepregistry.get_pattern(step_c)

    # then
    assert pattern_a == "step_pattern_a"
    assert pattern_b == "step_pattern_b"
    assert pattern_c == "Unknown"



# -*- coding: utf-8 -*-

"""
    Providing radish core functionality like the feature file Runner.
"""

from __future__ import unicode_literals

from random import shuffle

from .terrain import world
from .scenariooutline import ScenarioOutline
from .scenarioloop import ScenarioLoop
from .stepmodel import Step


class Runner(object):
    """
        Represents a class which is able to run features.
    """

    def handle_exit(func):  # pylint: disable=no-self-argument
        """
            Handles an runner exit
        """

        def _decorator(self, *args, **kwargs):
            """
                Actual decorator
            """
            if self._required_exit:  # pylint: disable=protected-access
                return 1

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return _decorator

    def call_hooks(model):  # pylint: disable=no-self-argument
        """
            Call hooks for a specific model
        """

        def _decorator(func):
            """
                The actual decorator
            """

            def _wrapper(self, model_instance, *args, **kwargs):
                """
                    Decorator wrapper
                """
                self._hooks.call(
                    "before", model, True, model_instance, *args, **kwargs
                )  # pylint: disable=protected-access
                try:
                    return func(self, model_instance, *args, **kwargs)
                finally:
                    self._hooks.call(
                        "after", model, False, model_instance, *args, **kwargs
                    )  # pylint: disable=protected-access

            return _wrapper

        return _decorator

    def __init__(self, hooks, early_exit=False, show_only=False):
        self._hooks = hooks
        self._early_exit = early_exit
        self._required_exit = False
        self._show_only = show_only

    @handle_exit
    @call_hooks("all")
    def start(self, features, marker):
        """
            Start running features

            :param list features: the features to run
            :param string marker: the marker for this run
        """
        if world.config.shuffle:
            shuffle(features)

        returncode = 0  # return code is set to 1 if any feature fails
        for feature in features:
            if not feature.has_to_run(world.config.scenarios):
                continue

            returncode |= self.run_feature(feature)

        if world.config.wip:
            # swap the state from 0 to 1
            # and the other way around
            returncode ^= 1

        return returncode

    @handle_exit
    @call_hooks("each_feature")
    def run_feature(self, feature):
        """
            Runs the given feature

            :param Feature feature: the feature to run
        """
        if world.config.shuffle:
            shuffle(feature.scenarios)

        returncode = 0
        for scenario in feature:
            if not scenario.has_to_run(world.config.scenarios):
                continue
            returncode |= self.run_scenario(scenario)

            if not isinstance(scenario, (ScenarioOutline, ScenarioLoop)):
                continue

            for sub_scenario in scenario.scenarios:
                returncode |= self.run_scenario(sub_scenario)
        return returncode

    @handle_exit
    @call_hooks("each_scenario")
    def run_scenario(self, scenario):
        """
            Runs the given scenario

            :param Scenario scenario: the scnenario to run
        """
        returncode = 0
        steps = scenario.all_steps if world.config.expand else scenario.steps
        for step in steps:
            if scenario.state == Step.State.FAILED:
                self.skip_step(step)
                continue

            returncode |= self.run_step(step)

            if step.state == step.State.FAILED and self._early_exit:
                self.exit()
                return 1
        return returncode

    @handle_exit
    @call_hooks("each_step")
    def run_step(self, step):
        """
            Runs the given step

            :param Step step: the step to run
        """
        if self._show_only:
            return 0

        if world.config.debug_steps:
            state = step.debug()
        else:
            state = step.run()

        return 1 if state == Step.State.FAILED else 0

    def skip_step(self, step):
        """
            Skips the given step

            :param Step step: the step to skip
        """
        self._hooks.call("before", "each_step", True, step)
        step.skip()
        self._hooks.call("after", "each_step", False, step)

    def exit(self):
        """
            Exits the runner
        """
        self._required_exit = True



# -*- coding: utf-8 -*-

"""
    This radish extension provides the functionality to write the feature file run to the console.
"""

from __future__ import unicode_literals
from __future__ import print_function

import sys
import colorful as cf

from radish.terrain import world
from radish.hookregistry import before, after
from radish.scenariooutline import ScenarioOutline
from radish.scenarioloop import ScenarioLoop
from radish.stepmodel import Step
from radish.extensionregistry import extension
from radish.compat import u


@extension
class DotOutputFormatter(object):
    """
    Output formatter in the dot style.
    """

    LOAD_IF = staticmethod(lambda config: config.formatter == "dots")
    LOAD_PRIORITY = 30

    STATE_SYMBOLS = {
        Step.State.PASSED: ".",
        Step.State.PENDING: "P",
        Step.State.UNTESTED: "U",
        Step.State.SKIPPED: "S",
        Step.State.FAILED: "F",
    }

    def __init__(self):
        before.each_feature(self.dot_formatter_before_each_feature)
        after.each_feature(lambda *args, **kwargs: sys.stdout.write("\n"))
        after.each_scenario(self.dot_formatter_after_each_scenario)
        after.each_step(self.dot_formatter_after_each_step)
        after.all(self.dot_formatter_failure_summary)

        self._failed_steps = []

    def dot_formatter_before_each_feature(self, feature):
        """
            Writes feature header to the console

            :param Feature feature: the feature to write to the console
        """
        output = cf.bold_black(feature.path) + ": "
        sys.stdout.write(u(output))

    def dot_formatter_after_each_scenario(self, scenario):
        """
            If the scenario is a ExampleScenario it will write the Examples header

            :param Scenario scenario: the scenario which was ran.
        """
        if isinstance(scenario, (ScenarioOutline, ScenarioLoop)):
            return

        sys.stdout.write(u(self.STATE_SYMBOLS[scenario.state]))

    def dot_formatter_after_each_step(self, step):
        if step.state == Step.State.FAILED:
            self._failed_steps.append(step)

    def dot_formatter_failure_summary(self, features, marker):
        """Output summary for failed Scenarios."""
        if not self._failed_steps:
            return

        output = "\n" + cf.bold_red("Failures:") + "\n"

        for step in self._failed_steps:
            output += "{}: {}\n    {}\n".format(
                step.path, step.parent.sentence, cf.red(step.sentence)
            )
            if world.config.with_traceback:
                output += "      {}\n".format(
                    "\n      ".join(
                        [
                            str(cf.red(l))
                            for l in step.failure.traceback.split("\n")[:-2]
                        ]
                    )
                )
            output += "      {}: {}\n\n".format(
                cf.bold_red(step.failure.name), cf.red(step.failure.reason)
            )

        sys.stdout.write(u(output + "\n"))



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import os
import re
import tempfile

import pytest

import colorful

import radish.testing.matches as matches


@pytest.fixture(scope="module", autouse=True)
def disable_colors():
    """
    Fixture to disable colors
    """
    colorful.disable()
    yield
    colorful.use_8_ansi_colors()


def test_unreasonable_min_coverage(capsys):
    """
    Test unreasonable minimum test coverage
    """
    # given
    min_coverage = 101
    expected_returncode = 3

    # when
    actual_returncode = matches.test_step_matches_configs(None, [], min_coverage)
    _, err = capsys.readouterr()

    # then
    assert actual_returncode == expected_returncode
    assert (
        err
        == "You are a little cocky to think you can reach a minimum coverage of 101.00%\n"
    )


def test_no_steps_found(mocker, capsys):
    """
    Test if basedir does not contain any Steps to test against
    """
    # given
    mocker.patch("radish.testing.matches.load_modules")
    expected_returncode = 4

    # when
    actual_returncode = matches.test_step_matches_configs(None, [])
    _, err = capsys.readouterr()

    # then
    assert actual_returncode == expected_returncode
    assert (
        err
        == "No step implementations found in [], thus doesn't make sense to continue"
    )


def test_empty_matches_config(mocker, capsys):
    """
    Test empty matches config file
    """
    # given
    expected_returncode = 5

    fd, tmpfile = tempfile.mkstemp()
    os.close(fd)

    mocker.patch("radish.testing.matches.load_modules")
    steps_mock = mocker.patch("radish.testing.matches.StepRegistry.steps")
    steps_mock.return_value = [1, 2]

    # when
    actual_returncode = matches.test_step_matches_configs([tmpfile], [])
    out, _ = capsys.readouterr()

    # then
    assert actual_returncode == expected_returncode
    assert out == "No sentences found in {0} to test against\n".format(tmpfile)


@pytest.mark.parametrize(
    "given_invalid_config, expected_error_msg",
    [
        (
            [{"sentence": None}],
            "You have to provide a sentence and the function name which should (not) be matched (should_match, should_not_match)",
        ),
        (
            [{"should_match": None}],
            "You have to provide a sentence and the function name which should (not) be matched (should_match, should_not_match)",
        ),
        (
            [{}],
            "You have to provide a sentence and the function name which should (not) be matched (should_match, should_not_match)",
        ),
        (
            [{"foo": None}],
            "The config attributes foo are invalid. Use only sentence, should_match, should_not_match, with_arguments",
        ),
    ],
    ids=[
        "test config with missing should_match function attribute",
        "test config with missing sentence attribute",
        "test empty config",
        "test invalid config items",
    ],
)
def test_step_matches_invalid_config(given_invalid_config, expected_error_msg):
    """
    Test match config without a sentence and a should_match attribute
    """
    # given & when
    with pytest.raises(ValueError) as exc:
        matches.test_step_matches(given_invalid_config, [])

    # then
    assert str(exc.value) == expected_error_msg


def test_sentence_no_step_match(capsys):
    """
    Test if sentence does not match any Step Pattern
    """
    # given
    steps = {}
    config = [{"sentence": "foo", "should_match": "bar"}]
    expected_returncode = (1, 0)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert "Expected sentence didn't match any step implementation" in out
    assert actual_returncode == expected_returncode


def test_sentence_match_wrong_step(capsys):
    """
    Test if sentence matched wrong step
    """
    # given
    def foo():
        pass

    steps = {"foo": foo}
    config = [{"sentence": "foo", "should_match": "bar"}]
    expected_returncode = (1, 0)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert "Expected sentence matched foo instead of bar" in out
    assert actual_returncode == expected_returncode


def test_sentence_argument_errors(capsys):
    """
    Test if sentence arguments do not match
    """
    # given
    def foo(step, foo, bar):
        pass

    steps = {re.compile(r"What (.*?) can (.*)"): foo}
    config = [
        {
            "sentence": "What FOO can BAR",
            "should_match": "foo",
            "with_arguments": [{"foo": "foooooooo"}, {"bar": "baaaaaaar"}],
        }
    ]
    expected_returncode = (1, 0)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert (
        'Expected argument "foo" with value "foooooooo" does not match value "FOO"'
        in out
    )
    assert (
        'Expected argument "bar" with value "baaaaaaar" does not match value "BAR"'
        in out
    )
    assert actual_returncode == expected_returncode


def test_sentence_not_match(capsys):
    """
    Test if sentence does not match if that's expected
    """
    # given
    def foo(step, foo, bar):
        pass

    steps = {re.compile(r"What (.*?) can (.*)"): foo}
    config = [{"sentence": "foo", "should_not_match": "foo"}]
    expected_returncode = (0, 1)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert actual_returncode == expected_returncode


def test_sentence_not_match_anything(capsys):
    """
    Test if sentence does not match any steps
    """
    # given
    def foo(step, foo, bar):
        pass

    steps = {re.compile(r"What (.*?) can (.*)"): foo}
    config = [{"sentence": "foo", "should_not_match": ""}]
    expected_returncode = (0, 1)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert actual_returncode == expected_returncode


def test_sentence_not_match_specific_step(capsys):
    """
    Test if sentence does not match specific step
    """
    # given
    def foo(step):
        pass

    def bar(step):
        pass

    steps = {re.compile(r"foo"): foo, re.compile(r"bar"): bar}
    config = [{"sentence": "foo", "should_not_match": "bar"}]
    expected_returncode = (0, 1)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert actual_returncode == expected_returncode


def test_sentence_not_match_but_does(capsys):
    """
    Test if sentence matched step but shouldn't
    """
    # given
    def foo():
        pass

    steps = {"foo": foo}
    config = [{"sentence": "foo", "should_not_match": "foo"}]
    expected_returncode = (1, 0)

    # when
    actual_returncode = matches.test_step_matches(config, steps)
    out, _ = capsys.readouterr()

    # then
    assert "Expected sentence did match foo but it shouldn't" in out
    assert actual_returncode == expected_returncode


@pytest.mark.parametrize(
    "given_actual_arguments, expected_messages",
    [
        (
            {"FOO": None},
            ["Expected argument \"foo\" is not in matched arguments ['FOO']"],
        ),
        ({"foo": 42}, ['Expected argument "foo" is of type "int" instead "str"']),
        (
            {"foo": "foo"},
            ['Expected argument "foo" with value "fooo" does not match value "foo"'],
        ),
    ],
    ids=["Missing argument", "Wrong type argument", "Wrong value argument"],
)
def test_checking_step_arguments_errors(given_actual_arguments, expected_messages):
    """
    Test step argument checking errors
    """
    # given
    expected_arguments = {"foo": "fooo"}

    # when
    messages = matches.check_step_arguments(expected_arguments, given_actual_arguments)

    # then
    assert messages == expected_messages


def test_checking_step_arguments():
    """
    Test sentence step arguments match
    """
    # given
    expected_arguments = {"foo": "foo", "bar": 42}

    actual_arguments = {"foo": "foo", "bar": 42}

    # when
    messages = matches.check_step_arguments(expected_arguments, actual_arguments)

    # then
    assert messages == []


def test_checking_step_arguments_directory():
    """
    Test sentence step argument directory match
    """
    # given
    expected_arguments = {"foobar": {"foo": "foo", "bar": "bar"}}

    actual_arguments = {"foobar": {"foo": "foo", "bar": "bar"}}

    # when
    messages = matches.check_step_arguments(expected_arguments, actual_arguments)

    # then
    assert messages == []



# -*- coding: utf-8 -*-

"""
    This module provides an extension to write all features, scenarios and steps to the syslog.
"""

from __future__ import unicode_literals

from radish.terrain import world
from radish.feature import Feature
from radish.hookregistry import before, after
from radish.extensionregistry import extension


@extension
class SyslogWriter(object):
    """
        Syslog Writer radish extension. This extension is only supported on
        systems where the Python standard library supports the system logger
        (syslog). For example, this extension works on UNIX and UNIX-like
        systems (Linux), but will not work on Windows.
    """

    OPTIONS = [
        ("--syslog", "log all of your features, scenarios, and steps to the syslog")
    ]
    LOAD_IF = staticmethod(lambda config: config.syslog)
    LOAD_PRIORITY = 40

    def __init__(self):
        # import syslog only if the extension got loaded
        # but not if the module got loaded.
        import syslog

        before.all(self.syslog_writer_before_all)
        before.each_feature(self.syslog_writer_before_each_feature)
        before.each_scenario(self.syslog_writer_before_each_scenario)
        before.each_step(self.syslog_writer_before_each_step)
        after.all(self.syslog_writer_after_all)
        after.each_feature(self.syslog_writer_after_each_feature)
        after.each_scenario(self.syslog_writer_after_each_scenario)
        after.each_step(self.syslog_writer_after_each_step)

    def get_scenario_feature(self, scenario):
        """
            Gets the scenarios feature
        """
        if not isinstance(scenario.parent, Feature):
            return scenario.parent.parent

        return scenario.parent

    def log(self, message):
        """
            Logs the given message to the syslog

            :param string message: the message to log
        """
        import syslog

        try:
            if isinstance(message, unicode):
                message = message.encode("utf8")
        except Exception:  # pylint: disable=broad-except
            pass
        finally:
            syslog.syslog(syslog.LOG_INFO, message)

    def syslog_writer_before_all(
        self, features, marker
    ):  # pylint: disable=unused-argument
        """
            Opens the syslog
        """
        import syslog

        syslog.openlog(b"radish")
        self.log("begin run {0}".format(marker))

    def syslog_writer_after_all(
        self, features, marker
    ):  # pylint: disable=unused-argument
        """
            Closes the syslog
        """
        import syslog

        self.log("end run {0}".format(marker))
        syslog.closelog()

    def syslog_writer_before_each_feature(self, feature):
        """
            Writes the feature to the syslog
        """
        self.log(
            "begin feature {0}:{1} {2}".format(
                world.config.marker, feature.id, feature.sentence
            )
        )

    def syslog_writer_after_each_feature(self, feature):
        """
            Writes the feature to the syslog
        """
        self.log(
            "end feature {0}:{1} {2}".format(
                world.config.marker, feature.id, feature.sentence
            )
        )

    def syslog_writer_before_each_scenario(self, scenario):
        """
            Writes the scenario to the syslog
        """
        self.log(
            "begin scenario {0}:{1}.{2} {3}".format(
                world.config.marker,
                self.get_scenario_feature(scenario).id,
                scenario.id,
                scenario.sentence,
            )
        )

    def syslog_writer_after_each_scenario(self, scenario):
        """
            Writes the scenario to the syslog
        """
        self.log(
            "end scenario {0}:{1}.{2} {3}".format(
                world.config.marker,
                self.get_scenario_feature(scenario).id,
                scenario.id,
                scenario.sentence,
            )
        )

    def syslog_writer_before_each_step(self, step):
        """
            Writes the step to the syslog
        """
        self.log(
            "begin step {0}:{1}.{2}.{3} {4}".format(
                world.config.marker,
                self.get_scenario_feature(step.parent).id,
                step.parent.id,
                step.id,
                step.sentence,
            )
        )

    def syslog_writer_after_each_step(self, step):
        """
            Writes the step to the syslog
        """
        self.log(
            "{0} step {1}:{2}.{3}.{4} {5}".format(
                step.state,
                world.config.marker,
                self.get_scenario_feature(step.parent).id,
                step.parent.id,
                step.id,
                step.sentence,
            )
        )



# -*- coding: utf-8 -*-

"""
This module provides a radish extension to make
coverage measurements.
"""

from __future__ import unicode_literals

import sys
import re

from radish.extensionregistry import extension
from radish.hookregistry import before, after
from radish.terrain import world
from radish.compat import StringIO
from radish.exceptions import RadishError


@extension
class CodeCoverage(object):
    """
        Code Coverage radish extension
    """

    OPTIONS = [
        ("--with-coverage", "enable code coverage"),
        ("--cover-packages=<cover_packages>", "specify source code package"),
        ("--cover-append", "append coverage data to previous collected data"),
        (
            "--cover-config-file=<cover_config_file>",
            "specify coverage config file [default: .coveragerc]",
        ),
        ("--cover-branches", "include branch coverage in report"),
        ("--cover-erase", "erase previously collected coverage data"),
        (
            "--cover-min-percentage=<cover_min_percentage>",
            "fail if the given minimum coverage percentage is not reached",
        ),
        (
            "--cover-html=<cover_html_dir>",
            "specify a directory where to store HTML coverage report",
        ),
        (
            "--cover-xml=<cover_xml_file>",
            "specify a file where to store XML coverage report",
        ),
    ]
    LOAD_IF = staticmethod(lambda config: config.with_coverage)
    LOAD_PRIORITY = 70

    def __init__(self):
        try:
            from coverage import Coverage
        except ImportError:
            raise RadishError(
                'if you want to use the code coverage you have to "pip install radish-bdd[coverage]"'
            )

        before.all(self.coverage_start)
        after.all(self.coverage_stop)

        if world.config.cover_packages:
            self.cover_packages = world.config.cover_packages.split(",")
        else:
            self.cover_packages = []

        self.coverage = None
        self.modules_on_init = set(sys.modules.keys())

    def coverage_start(self, features, marker):
        """
        Start the coverage measurement
        """
        from coverage import Coverage

        # if no explicit modules are specified we just
        # use the ones loaded from radish's basedir.
        # During the plugin init the basedir modules are
        # not loaded yet, but they are during the start method.
        # Thus, we are safe to consider the difference between the
        # two for coverage measurement.
        if not self.cover_packages:
            source = list(set(sys.modules.keys()).difference(self.modules_on_init))
        else:
            source = self.cover_packages

        self.coverage = Coverage(
            source=source,
            config_file=world.config.cover_config_file,
            branch=world.config.cover_branches,
        )
        if world.config.cover_erase:
            self.coverage.combine()
            self.coverage.erase()

        if world.config.cover_append:
            self.coverage.load()
        self.coverage.start()

    def coverage_stop(self, features, marker):
        """
        Stop the coverage measurement
        and create report
        """
        self.coverage.stop()
        self.coverage.combine()
        self.coverage.save()
        self.coverage.report(file=sys.stdout)

        if world.config.cover_html:
            self.coverage.html_report(directory=world.config.cover_html)

        if world.config.cover_xml:
            self.coverage.xml_report(outfile=world.config.cover_xml)

        if world.config.cover_min_percentage:
            report = StringIO()
            self.coverage.report(file=report)
            match = re.search(r"^TOTAL\s+(.*)$", report.getvalue(), re.MULTILINE)
            if not match:
                raise RadishError("Failed to find total percentage in coverage report")

            total_percentage = int(match.groups()[0].split()[-1][:-1])
            if total_percentage < int(world.config.cover_min_percentage):
                raise RadishError(
                    "Failed to reach minimum expected coverage of {0}% (reached: {1}%)".format(
                        world.config.cover_min_percentage, total_percentage
                    )
                )



# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import os
import json

from radish import given, when, then, world
from radish.extensions.cucumber_json_writer import CucumberJSONWriter


@given("I have a step")
def have_a_step(step):
    "Given I have a step"
    pass


@when("I do something")
def do_something(step):
    "When I do something"
    pass


@then("I expect something")
def expect_something(step):
    "Then I expect something"
    pass


@given("I have the number {number:d}")
def have_number(step, number):
    "Given I have the number <n>"
    if not hasattr(step.context, "numbers"):
        step.context.numbers = []

    step.context.numbers.append(number)


@when("I add them up")
def sum_numbers(step):
    "When I add them up"
    step.context.sum = sum(step.context.numbers)


@when("I add them up with failure")
def sum_numbers(step):
    "When I add them up with failure"
    assert False, "Unable to add numbers: {0}".format(step.context.numbers)


@when("I subtract them")
def subtract_numbers(step):
    "When I subtract them up"
    difference = step.context.numbers[0]
    for n in step.context.numbers[1:]:
        difference -= n
    step.context.difference = difference


@then("I expect the sum to be {expected_sum:d}")
def expect_sum(step, expected_sum):
    "Then I expect the sum to be <n>"
    assert (
        step.context.sum == expected_sum
    ), "The expected sum {0} does not match actual sum {1}".format(
        expected_sum, step.context.sum
    )


@then("I expect the difference to be {expected_diff:d}")
def expect_sum(step, expected_diff):
    "Then I expect the difference to be <n>"
    assert (
        step.context.difference == expected_diff
    ), "The expected difference {0} does not match actual difference {1}".format(
        expected_diff.step.context.difference
    )


@given("I have an instable function")
def have_instable_function(step):
    "Given I have an instable function"
    pass


@when("I execute it")
def execute_instable_function(step):
    "When I execute it"
    pass


@then("I expect it to pass")
def expect_instable_function_pass(step):
    "Then I expect it to pass"
    pass


@given("I have the following heros")
def have_heros(step):
    "Given I have the following heros"
    step.context.heros = step.table


@when("I capitalize their first name")
def cap_first_name(step):
    "When I capitalize their first name"
    for hero in step.context.heros:
        hero["firstname"] = hero["firstname"].upper()


@then("I have the following names")
def have_names(step):
    "Then I have the following names"
    assert list(x["firstname"] for x in step.context.heros) == list(
        x["cap_heroname"] for x in step.table
    )


@given("I have the following quote")
def have_quote(step):
    "Given I have the following quote"
    step.context.quote = step.text


@when("I look for it's author")
def lookup_author(step):
    "When I look for it's author"
    step.context.author = "Shakespeare"


@then("I will find {:S}")
def expect_author(step, author):
    "Then I will find <author>"
    assert step.context.author == author


@when("I embed a text {test_text:QuotedString}")
def embed_a_text(step, test_text):
    'When I embed a text "<test_text>"'
    step.embed(test_text)
    step.context.step_with_embedded_data = step


@then("step with embedded text should have following embedded data")
def embed_a_text(step):
    "Then step with embedded text should have following embedded data"
    assert hasattr(
        step.context, "step_with_embedded_data"
    ), "step_embeddings is missing in context - please check if step with text embedding has been executed"
    test_step_embeddings = step.context.step_with_embedded_data.embeddings
    for embeddings in step.table:
        assert embeddings in test_step_embeddings, "{0} not found in {1}".format(
            embeddings, test_step_embeddings
        )


@when("generate cucumber report")
def generate_cucumber_report(step):
    cjw = CucumberJSONWriter()
    cjw.generate_ccjson([step.parent.parent], None)


@then("genreated cucumber json equals to {expected_json_file:QuotedString}")
def proper_cucumber_json_is_generated(step, expected_json_file):
    def remove_changing(d):
        return {k: v for k, v in d.items() if k not in ["duration", "uri"]}

    with open(world.config.cucumber_json, "r") as f_cucumber_json:
        cucumber_json = json.load(f_cucumber_json, object_hook=remove_changing)
    json_file_path = os.path.join(
        os.path.dirname(step.path), "..", "output", expected_json_file
    )
    with open(json_file_path, "r") as f_expected_cucumber_json:
        expected_cucumber_json = json.load(
            f_expected_cucumber_json, object_hook=remove_changing
        )
    assert cucumber_json == expected_cucumber_json


@when("YAML specification is set to")
def yaml_specification_is_set_to(step):
    step.context.doc_text = step.text

@then("YAML specification contains proper data")
def yaml_specification_contains_correct_data(step):
    expected_data = """version: '3'
services:
  webapp:
    build: ./dir"""
    assert step.context.doc_text == expected_data, '"{}" != "{}"'.format(step.context.doc_text, expected_data)


# -*- coding: utf-8 -*-

"""
    radish
    ~~~~~~

    Behavior Driven Development tool for Python - the root from red to green

    Copyright: MIT, Timo Furrer <tuxtimo@gmail.com>
"""

import re

import pytest

from parse_type.cfparse import Parser

import radish.matcher as matcher
import radish.exceptions as errors


@pytest.mark.parametrize(
    "regex, string, expected_args, expected_kwargs",
    [
        (r"Given I have the number (\d+)", "Given I have the number 5", ("5",), {}),
        (
            r"Given I have the number (?P<number>\d+)",
            "Given I have the number 5",
            ("5",),
            {"number": "5"},
        ),
        (
            r"Given I have the number (?P<number>\d+) and (\d+)",
            "Given I have the number 4 and 2",
            ("4", "2"),
            {"number": "4"},
        ),
    ],
    ids=[
        "Regex with unnamed groups",
        "Regex with named groups",
        "Regex with unnamed and named groups",
    ],
)
def test_regex_step_arguments_object(regex, string, expected_args, expected_kwargs):
    """
    Test functionality of RegexStepArguments object
    """
    # given
    match = re.search(regex, string)
    args = matcher.RegexStepArguments(match)

    # when
    actual_args, actual_kwargs = args.evaluate()

    # then
    assert actual_args == expected_args
    assert actual_kwargs == expected_kwargs


@pytest.mark.parametrize(
    "parse_pattern, string, expected_args, expected_kwargs",
    [
        ("Given I have the number {:d}", "Given I have the number 5", (5,), {}),
        (
            "Given I have the number {number:d}",
            "Given I have the number 5",
            tuple(),
            {"number": 5},
        ),
        (
            "Given I have the number {number:d} and {:d}",
            "Given I have the number 4 and 2",
            (2,),
            {"number": 4},
        ),
    ],
    ids=[
        "Parse Pattern with unnamed groups",
        "Parse Pattern with named groups",
        "Parse Pattern with unnamed and named groups",
    ],
)
def test_parse_step_arguments_object(
    parse_pattern, string, expected_args, expected_kwargs
):
    """
    Test functionality of ParseStepArguments object
    """
    # given
    parser = Parser(parse_pattern)
    match = parser.search(string, evaluate_result=False)
    args = matcher.ParseStepArguments(match)

    # when
    actual_args, actual_kwargs = args.evaluate()

    # then
    assert actual_args == expected_args
    assert actual_kwargs == expected_kwargs


@pytest.mark.parametrize(
    "given_sentence, given_steps, expected_argument_match_type, expected_func_match",
    [
        (
            "Given I have a number",
            {re.compile("Given I have a number"): 1},
            matcher.RegexStepArguments,
            1,
        ),
        (
            "Given I have a number to test",
            {re.compile("Given I have a number"): 1},
            matcher.RegexStepArguments,
            1,
        ),
        (
            "Given I have a number",
            {
                re.compile("Given I have some number"): 1,
                re.compile("Given I have a number"): 2,
            },
            matcher.RegexStepArguments,
            2,
        ),
        (
            "Given I have a number",
            {re.compile("I have a number"): 1, re.compile("Given I have a number"): 2},
            matcher.RegexStepArguments,
            2,
        ),
        (
            "Given I have the number 5",
            {
                re.compile(r"I have the number (\d+)"): 1,
                re.compile(r"Given I have the number (\d+)"): 2,
            },
            matcher.RegexStepArguments,
            2,
        ),
        (
            "Given I have the number 5 which I use to test",
            {
                re.compile(
                    r"Given I have the number (\d+) which I use to do some stuff"
                ): 1,
                re.compile(r"Given I have the number (\d+) which I use"): 2,
            },
            matcher.RegexStepArguments,
            2,
        ),
        # Parse Argument Matches
        (
            "Given I have a number",
            {"Given I have a number": 1},
            matcher.ParseStepArguments,
            1,
        ),
        (
            "Given I have a number to test",
            {"Given I have a number": 1},
            matcher.ParseStepArguments,
            1,
        ),
        (
            "Given I have a number",
            {"Given I have some number": 1, "Given I have a number": 2},
            matcher.ParseStepArguments,
            2,
        ),
        (
            "Given I have a number",
            {"I have a number": 1, "Given I have a number": 2},
            matcher.ParseStepArguments,
            2,
        ),
        (
            "Given I have the number 5",
            {"I have the number {:d}": 1, "Given I have the number {:d}": 2},
            matcher.ParseStepArguments,
            2,
        ),
        (
            "Given I have the number 5 which I use to test",
            {
                "Given I have the number {:d} which I use to do some stuff": 1,
                "Given I have the number {:d} which I use": 2,
            },
            matcher.ParseStepArguments,
            2,
        ),
    ],
    ids=[
        "Regex: match simple step with single candidate",
        "Regex: match simple step with single candidate which is not a perfect match",
        "Regex: match simple step with one matching candidate",
        "Regex: match simple step with multiple candidate - best match",
        "Regex: match step with args with multiple candidate - best match",
        "Regex: match step with args with multiple candidate - closest match",
        # Parse Argument Matches
        "Parse: match simple step with single candidate",
        "Parse: match simple step with single candidate which is not a perfect match",
        "Parse: match simple step with one matching candidate",
        "Parse: match simple step with multiple candidate - best match",
        "Parse: match step with args with multiple candidate - best match",
        "Parse: match step with args with multiple candidate - closest match",
    ],
)
def test_match_sentence_with_steps(
    given_sentence, given_steps, expected_argument_match_type, expected_func_match
):
    """
    Test matching a sentence with given Steps
    """
    # given & when
    match = matcher.match_step(given_sentence, given_steps)

    # then
    assert isinstance(match.argument_match, expected_argument_match_type)
    assert match.func == expected_func_match


@pytest.mark.parametrize(
    "given_sentence, given_steps",
    [
        ("Given I have the number", {re.compile("Given I have a number"): 1}),
        ("Given I have the number foo", {re.compile(r"Given I have number (\d+)"): 1}),
        ("Given I have the number", {"Given I have a number": 1}),
        ("Given I have the number foo", {"Given I have number {:d}": 1}),
    ],
    ids=[
        "Regex: simple not matching step",
        "Regex: not matching step with arguments",
        "Parse: simple not matching step",
        "Parse: not matching step with arguments",
    ],
)
def test_no_step_match(given_sentence, given_steps):
    """
    Test failing to match a sentence with given Steps
    """
    # given & when
    match = matcher.match_step(given_sentence, given_steps)

    # then
    assert match is None


def test_invalid_parse_pattern():
    """
    Test failure for invalid Parse pattern
    """
    # given
    invalid_pattern = "Given I have the number {:d {}"

    # when
    with pytest.raises(errors.StepPatternError) as exc:
        matcher.match_step("Given I have the number 5", {invalid_pattern: int})

    assert str(exc.value).startswith(
        "Cannot compile pattern 'Given I have the number {:d {}' of step 'int': "
    )


@pytest.mark.parametrize(
    "given_sentence, given_steps, expected_func, expected_args, expected_kwargs",
    [
        (
            "Given I have a number",
            {re.compile(r"Given I have a number"): 1},
            1,
            tuple(),
            {},
        ),
        (
            "Given I have the number 4 and 2",
            {re.compile(r"Given I have the number (?P<number>\d+) and (\d+)"): 1},
            1,
            ("4", "2"),
            {"number": "4"},
        ),
        ("Given I have a number", {"Given I have a number": 1}, 1, tuple(), {}),
        (
            "Given I have the number 4 and 2",
            {"Given I have the number {number:d} and {:d}": 1},
            1,
            (2,),
            {"number": 4},
        ),
    ],
    ids=[
        "Regex: simple step merge",
        "Regex: step merge with arguments",
        "Parse: simple step merge",
        "Parse: step merge with arguments",
    ],
)
def test_merging_step(
    given_sentence, given_steps, expected_func, expected_args, expected_kwargs, mocker
):
    """
    Test merging a Step with registered Step functions
    """
    # given
    step_to_merge = mocker.MagicMock(
        context_sensitive_sentence=given_sentence,
        definition_func=None,
        argument_match=None,
    )

    # when
    matcher.merge_step(step_to_merge, given_steps)
    actual_args, actual_kwargs = step_to_merge.argument_match.evaluate()

    # then
    assert step_to_merge.definition_func == expected_func
    assert actual_args == expected_args
    assert actual_kwargs == expected_kwargs


@pytest.mark.parametrize(
    "given_sentence, given_steps",
    [
        ("Given I have the number", {re.compile("Given I have a number"): 1}),
        ("Given I have the number foo", {re.compile(r"Given I have number (\d+)"): 1}),
        ("Given I have the number", {"Given I have a number": 1}),
        ("Given I have the number foo", {"Given I have number {:d}": 1}),
    ],
    ids=[
        "Regex: simple not matching step",
        "Regex: not matching step with arguments",
        "Parse: simple not matching step",
        "Parse: not matching step with arguments",
    ],
)
def test_no_step_definition_merge(given_sentence, given_steps, mocker):
    """
    Test failure when no Step definition function can be found during Step merge
    """
    # given
    step_to_merge = mocker.MagicMock(
        context_sensitive_sentence=given_sentence,
        definition_func=None,
        argument_match=None,
    )

    # when
    with pytest.raises(errors.StepDefinitionNotFoundError) as exc:
        matcher.merge_step(step_to_merge, given_steps)

    # then
    assert str(exc.value).startswith("Cannot find step definition for step")



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

import os
import re
import codecs
from setuptools import setup, find_packages


def read_metafile(path):
    """
    Read contents from given metafile
    """
    with codecs.open(path, "rb", "utf-8") as f:
        return f.read()


def get_meta(name):
    """
    Get some metdata with the give name from the
    Meta file
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=name.upper()),
        __META_DATA__,
        re.M,
    )

    if not meta_match:
        raise RuntimeError("Unable to find __{0}__ string.".format(name))
    return meta_match.group(1)


__META_FILE__ = os.path.join("radish", "__init__.py")
__META_DATA__ = read_metafile(__META_FILE__)
__README_CONTENTS__ = open("README.md").read()


# mandatory requirements for the radish base features
requirements = [
    "docopt",
    "pysingleton",
    "colorful>=0.3.11",
    "tag-expressions>=1.0.0",
    "parse_type>0.4.0",
    "humanize",
]


# optional requirements used by extensions which are
# disabled by default
extra_requirements = {
    "bddxml": ["lxml"],
    "ipython-debugger": ["ipython"],
    "coverage": ["coverage"],
    "testing": ["PyYAML"],
}

setup(
    name="radish-bdd",
    version=get_meta("version"),
    license=get_meta("license"),
    description=get_meta("description"),
    long_description=__README_CONTENTS__,
    long_description_content_type="text/markdown",
    author=get_meta("author"),
    author_email=get_meta("author_email"),
    maintainer=get_meta("author"),
    maintainer_email=get_meta("author_email"),
    platforms=["Linux", "Windows", "MAC OS X"],
    url=get_meta("url"),
    download_url=get_meta("download_url"),
    bugtrack_url=get_meta("bugtrack_url"),
    packages=find_packages(),
    package_data={"": ["radish/languages/*", "*.md"]},
    install_requires=requirements,
    extras_require=extra_requirements,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "radish = radish.main:main",
            "radish-test = radish.testing.__main__:main [testing]",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: OS Independent",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Microsoft :: Windows :: Windows 7",
        "Operating System :: Microsoft :: Windows :: Windows XP",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation",
        "Topic :: Education :: Testing",
        "Topic :: Software Development",
        "Topic :: Software Development :: Testing",
    ],
)
